type arbre = {d: int; no: string; ne: string; so: string; se: string};;

let vide = {d = 0; no = ""; ne = ""; so = ""; se = ""};;


(* Initialisation de grandeurs globales. *)
print_string "Degre : ";;
let degre = read_int ();;
print_string "Nombre de generation : ";;
let nbr = read_int ();;
print_string "Vide : 0\nPlaneur : 1\nRandom : 2\n\nVotre choix : ";;
let choix = read_int ();;


(* On cre diffrentes tables de hachage, contenant chacune des strings de mme longueur. *)
let tab = Array.make (degre - 1) (Hashtbl.create 1);;
for k = 0 to degre - 2 do tab.(k) <- Hashtbl.create (1 lsl (k + 2)) done;;

let clear_all v =
  for k = 0 to degre - 2 do
    Hashtbl.clear v.(k) done;;


(* Les trois programmes suivant codent le Jeu de la Vie. *)
let somme m (i, j) =
  m.(i - 1).(j - 1) + m.(i - 1).(j) + m.(i - 1).(j + 1) + m.(i).(j + 1) + m.(i + 1).(j + 1) + m.(i + 1).(j) + m.(i + 1).(j - 1) + m.(i).(j - 1);;

let suivant cel voisin =
  if (cel = 0) && (voisin = 3) || (cel = 1) && (voisin = 2 || voisin = 3) then 1
  else 0;;

let calcul m =
  let no = string_of_int (suivant (m.(1).(1)) (somme m (1, 1)));
  and ne = string_of_int (suivant (m.(1).(2)) (somme m (1, 2)));
  and so = string_of_int (suivant (m.(2).(1)) (somme m (2, 1)));
  and se = string_of_int (suivant (m.(2).(2)) (somme m (2, 2))) in
    no ^ ne ^ so ^ se;;

(* Cette fonction calcul les cas de base une fois pour toute. On peut crire dans un fichier si besion est. *)
let base = Array.make 65536 "";;
let ind = ref 0 in
  for a = 0 to 1 do
    for b = 0 to 1 do
      for c = 0 to 1 do
        for d = 0 to 1 do
          for e = 0 to 1 do
            for f = 0 to 1 do
              for g = 0 to 1 do
                for h = 0 to 1 do
                  for i = 0 to 1 do
                    for j = 0 to 1 do
                      for k = 0 to 1 do
                        for l = 0 to 1 do
                          for m = 0 to 1 do
                            for n = 0 to 1 do
                              for o = 0 to 1 do
                                for p = 0 to 1 do
                                  base.(!ind) <- (calcul [|[|a; b; e; f|]; [|c; d; g; h|]; [|i; j; m; n|]; [|k; l; o; p|]|]);
                                  ind := !ind + 1
                                done
                              done
                            done
                          done
                        done
                      done
                    done
                  done
                done
              done
            done
          done
        done
      done
    done
  done;;


(* Renvoie le string associ  un arbre. Ce string fait aussi office de clef de hachage. *)
let nomme a =
  a.no ^ a.ne ^ a.so ^ a.se;;

(* Convertie un string en arbre. *)
let coupe d mot =
  let p = 1 lsl (2 * (d - 1))
  and q = 1 lsl (2 * d - 1) in
    {d = d;
      no = String.sub mot 0 p;
      ne = String.sub mot p p;
      so = String.sub mot q p;
      se = String.sub mot (p lor q) p};;

(*  deux arbres a1 a2 renvoie l'arbre au centre de la concatnation gauche/droite de a1 et a2 ie  a1 et a2 renvoie {a1.ne.se; a2.no.so; a1.se.ne; a2.so.no}. *)
let horizontal a1 a2 long =
  let no = String.sub a1 (7 * long) long;
  and ne = String.sub a2 (2 * long) long;
  and so = String.sub a1 (13 * long) long;
  and se = String.sub a2 (8 * long) long in
    no ^ ne ^ so ^ se;;

(*  deux arbres a1 a2 renvoie l'arbre au centre de la concatnation haut/bas de a1 et a2 ie  a1 et a2 renvoie {a1.so.se; a1.se.so; a2.no.ne; a2.ne.no}. *)
let vertical a1 a2 long =
  let no = String.sub a1 (11 * long) long;
  and ne = String.sub a1 (14 * long) long;
  and so = String.sub a2 (1 * long) long;
  and se = String.sub a2 (4 * long) long in
    no ^ ne ^ so ^ se;;
(*  un arbre a renvoie son arbre au centre ie  a renvoie {a.no.se; a.ne.so; a.so.ne; a.se.no}. *)
let centre a long =
  let no = String.sub a (3 * long) long;
  and ne = String.sub a (6 * long) long;
  and so = String.sub a (9 * long) long;
  and se = String.sub a (12 * long) long in
    no ^ ne ^ so ^ se;;

(* Applique deux fois la fonction prcdente  un arbre a. *)
let centrecentre a long =
  let no = String.sub a.no (15 * long) long;
  and ne = String.sub a.ne (10 * long) long;
  and so = String.sub a.so (5 * long) long;
  and se = String.sub a.se 0 long in
    no ^ ne ^ so ^ se;;


(* Dcoupe un arbre a de degr d en 9 sous-arbres de degr d-2 puis les recombine en 4 arbres de degr d-1 dont la concatnation de leur application  h donne la gnration suivante de l'arbre au centre de a.
La fonction h s'occupe de la gestion des Tables de Hachage et de la rcursivit. *)
let rec traite a = match a.d with
    | 1 -> ""
    | 2 -> base.(int_of_string ("0b" ^ (nomme a)))
    | d -> let long = (1 lsl (2*(d - 3))) in
          let a00 = centre a.no long
          and a01 = horizontal a.no a.ne long
          and a02 = centre a.ne long
          and a10 = vertical a.no a.so long
          and a11 = centrecentre a long
          and a12 = vertical a.ne a.se long
          and a20 = centre a.so long
          and a21 = horizontal a.so a.se long
          and a22 = centre a.se long in
            let ano = a00 ^ a01 ^ a10 ^ a11
            and ane = a01 ^ a02 ^ a11 ^ a12
            and aso = a10 ^ a11 ^ a20 ^ a21
            and ase = a11 ^ a12 ^ a21 ^ a22 in
              (h (d - 1) ano) ^ (h (d - 1) ane) ^ (h (d - 1) aso) ^ (h (d - 1) ase);
and h d mot =
  let m = try Hashtbl.find tab.(d - 2) mot with
      Not_found -> let a = traite (coupe d mot) in Hashtbl.add tab.(d - 2) mot a; a
  in m;;

(* Supprime les effets de bord en considrant le plan comme un tor et renvoie la gnration de cellules suivante. *)
let generation mot =
  let a = coupe degre mot in
    let ano = coupe (degre - 1) (a.no)
    and ane = coupe (degre - 1) (a.ne)
    and aso = coupe (degre - 1) (a.so)
    and ase = coupe (degre - 1) (a.se) in
      let a00 = (ase.se) ^ (aso.so) ^ (ane.ne) ^ (ano.no)
      and a01 = (aso.se) ^ (ase.so) ^ (ano.ne) ^ (ane.no)
      and a10 = (ane.se) ^ (ano.so) ^ (ase.ne) ^ (aso.no)
      and a11 = (ano.se) ^ (ane.so) ^ (aso.ne) ^ (ase.no) in
        let a0 = a00 ^ a01 ^ a10 ^ a11
        and a1 = a01 ^ a00 ^ a11 ^ a10
        and a2 = a10 ^ a11 ^ a00 ^ a01
        and a3 = a11 ^ a10 ^ a01 ^ a00 in
          {d = degre; no = h degre a0; ne = h degre a1; so = h degre a2; se = h degre a3};;

(* Calcul n gnration et les incrit dans un fichier. *)
let multigen mot n =
  let canal_out = open_out_bin "/Users/Benjamin/Desktop/test.txt" in
    let rec auxi m = function
        | 0 -> ()
        | n -> let s = nomme (generation m) in output_value canal_out s; auxi s (n - 1) in
      auxi mot n;
      close_out canal_out;;


(* Les quatres fonctions suivantes permettent de choisir entre diffrentes configurations test. *)
let random d =
  let rec auxi mot = function
      | 0 -> mot
      | n -> let a = Random.int 2 in auxi (mot ^ (string_of_int a)) (n - 1) in
    auxi "" (1 lsl (2 * d));;

let mat_nulle d = String.make (1 lsl (2 * d)) '0';;

let planeur d =
  let rec auxi = function
      | 0 -> failwith "Degre insuffisant"
      | 1 -> failwith "Degre insuffisant"
      | 2 -> "1100101001000000"
      | n -> let mot = auxi (n - 1) and nulle = mat_nulle (n - 1) in match (n mod 2) with
              | 0 -> mot ^ nulle ^ nulle ^ mot
              | _ -> nulle ^ mot ^ mot ^ nulle in
    auxi d;;

let choisir = function
    | 0 -> mat_nulle degre
    | 1 -> planeur degre
    | _ -> random degre;;


clear_all tab;;
multigen (choisir choix) nbr;;